import numpy as np
import gegd.parameter_processing.density_transforms as dtf
import time

from mpi4py import MPI
comm = MPI.COMM_WORLD

class optimizer:
    def __init__(self,
                 Nx,
                 Ny,
                 Nswarm,
                 symmetry,
                 periodic,
                 padding,
                 maxiter,
                 high_fidelity_setting,
                 brush_size,
                 upsample_ratio=1,
                 beta_proj=8,
                 brush_shape='circle',
                 cost_obj=None,
                 Nthreads=1,
                 ):
    
        self.Nx = Nx
        self.Ny = Ny
        self.Nswarm = Nswarm
        self.symmetry = symmetry
        self.periodic = periodic
        self.padding = padding
        self.maxiter = maxiter
        self.beta_proj = beta_proj
        self.brush_size = brush_size
        self.brush_shape = brush_shape
        self.upsample_ratio = upsample_ratio
        self.sigma_filter = brush_size/2/np.sqrt(2)
        self.high_fidelity_setting = high_fidelity_setting
        self.cost_obj = cost_obj
        self.Nthreads = Nthreads
        
        # Get Number of Independent Parameters
        if symmetry == 0:
            self.Ndim = Nx*Ny
           
        elif symmetry == 1:
            self.Ndim = int(np.floor(Nx/2 + 0.5)*Ny)
        
        elif symmetry == 2:
            self.Ndim = int(np.floor(Nx/2 + 0.5)*np.floor(Ny/2 + 0.5))
        
        elif symmetry == 4:
            self.Ndim = int(np.floor(Nx/2 + 0.5)*(np.floor(Nx/2 + 0.5) + 1)/2)

    def get_loss_swarm(self, x):
        # Get Brush Binarized Densities ------------------------------------------------------------
        x_bin = dtf.binarize(x, self.symmetry, self.periodic, self.Nx, self.Ny, self.brush_size, self.brush_shape, self.beta_proj, self.sigma_filter,
                             upsample_ratio=self.upsample_ratio, padding=self.padding, Nthreads=self.Nthreads)

        # Sample Modified Cost Function --------------------------------------------------------------
        self.cost_obj.set_accuracy(self.high_fidelity_setting)
        
        f_swarm = np.zeros(self.Nswarm)
        for n in range(self.Nswarm):
            f_swarm[n] = self.cost_obj.get_cost(x_bin[n,:], get_grad=False)
        
        return f_swarm, x_bin

    def PSO(self,
            lb,
            ub,
            coeff_cognitive=1.49,
            coeff_social=1.49,
            coeff_inertia=0.9,
            velocity=None,
            x=None,
            pbest_loss=None,
            pbest_x=None,
            pbest_x_bin=None,
            ):
    
        vmax = ub - lb
        if velocity is None and x is None:
            #Latin Hypercube Particle Initialization
            increment = (ub - lb)/self.Nswarm

            x = np.zeros((self.Nswarm, self.Ndim))
            velocity = np.zeros((self.Nswarm, self.Ndim))
            for nd in range(self.Ndim):
                swarm_ind = np.random.permutation(self.Nswarm).reshape(self.Nswarm)
                if lb[nd] == ub[nd]:
                    x[:,nd] = lb[nd]*np.ones(self.Nswarm)
                    velocity[:,nd] = np.zeros(self.Nswarm)
                else:
                    x[:,nd] = lb[nd] + (swarm_ind + 1)*increment[nd] - increment[nd]*np.random.rand(self.Nswarm)
                    velocity[:,nd] = -vmax[nd] + 2*vmax[nd]*np.random.rand(self.Nswarm)
    
        if pbest_loss is None and pbest_x is None:
            # Initialize pbest
            pbest_loss = np.zeros(self.Nswarm)
            pbest_x = np.zeros((self.Nswarm, self.Ndim))
    
        while True:
            t1 = time.time()
            if self.n_iter > self.maxiter:
                break
        
            loss_swarm, x_bin = self.get_loss_swarm(x)
            
            #Particle Best
            if self.n_iter == 0:
                pbest_loss = loss_swarm.copy()
                pbest_x = x.copy()
                pbest_x_bin = x_bin.copy()
            else:
                pbest_mask = loss_swarm < pbest_loss
                pbest_loss[pbest_mask] = loss_swarm[pbest_mask]
                pbest_x[pbest_mask,:] = x[pbest_mask,:]
                pbest_x_bin[pbest_mask,:] = x_bin[pbest_mask,:]
            
            #Global Best
            gbest_ind = np.argmin(pbest_loss)
            if self.gbest_x_hist is None:
                self.gbest_loss_hist = np.append(self.gbest_loss_hist, pbest_loss[gbest_ind])
                self.gbest_x_hist = pbest_x[gbest_ind,:].copy()
                self.gbest_x_binary_hist = pbest_x_bin[gbest_ind,:].copy()
            else:
                if self.gbest_loss_hist[-1] > pbest_loss[gbest_ind]:
                    self.gbest_loss_hist = np.append(self.gbest_loss_hist, pbest_loss[gbest_ind])
                    self.gbest_x_hist = np.vstack((self.gbest_x_hist, pbest_x[gbest_ind,:]))
                    self.gbest_x_binary_hist = np.vstack((self.gbest_x_binary_hist, pbest_x_bin[gbest_ind,:]))
                else:
                    self.gbest_loss_hist = np.append(self.gbest_loss_hist, self.gbest_loss_hist[-1])
                    if self.gbest_x_hist.ndim == 1:
                        self.gbest_x_hist = np.vstack((self.gbest_x_hist, self.gbest_x_hist))
                        self.gbest_x_binary_hist = np.vstack((self.gbest_x_binary_hist, self.gbest_x_binary_hist))
                    else:
                        self.gbest_x_hist = np.vstack((self.gbest_x_hist, self.gbest_x_hist[-1,:]))
                        self.gbest_x_binary_hist = np.vstack((self.gbest_x_binary_hist, self.gbest_x_binary_hist[-1,:]))

            t2 = time.time()
            t_rem = (t2 - t1)*(self.maxiter - self.n_iter + 1)/3600

            if comm.rank == 0:
                print('    | %12d | %12.5f | %12.5f | %12.5f |   %5.2f   |' %(self.n_iter, np.mean(pbest_loss), np.std(pbest_loss), self.gbest_loss_hist[-1], t_rem), flush=True)
            self.save_data(coeff_inertia=coeff_inertia,
                           velocity=velocity,
                           x=x,
                           pbest_loss=pbest_loss,
                           pbest_x=pbest_x,
                           pbest_x_bin=pbest_x_bin)

            #Parameter Updates
            if self.gbest_loss_hist.size >= 5 and self.gbest_loss_hist[-1] == self.gbest_loss_hist[-5]:
                coeff_inertia *= 0.95
            
            #Velocity Updates
            if self.gbest_x_hist.ndim == 1:
                velocity = coeff_inertia*velocity + coeff_cognitive*np.random.rand()*(pbest_x - x) + coeff_social*np.random.rand()*(self.gbest_x_hist[np.newaxis,:] - x)
            else:
                velocity = coeff_inertia*velocity + coeff_cognitive*np.random.rand()*(pbest_x - x) + coeff_social*np.random.rand()*(self.gbest_x_hist[-1,:][np.newaxis,:] - x)
            
            vmax_violation = np.abs(velocity) - vmax[np.newaxis,:]
            vmax_violation_mask = np.sum(np.abs(velocity) > vmax[np.newaxis,:], axis=1).astype(bool)
            vmax_violation_dim = np.argmax(vmax_violation, axis=1)[vmax_violation_mask]
            velocity[vmax_violation_mask,:] *= (vmax[vmax_violation_dim]/np.max(np.abs(velocity[vmax_violation_mask,:]), axis=1))[:,np.newaxis]
            
            #Craziness
            N_cr = np.random.permutation(self.Nswarm)[:int(self.Nswarm/10)]
            if np.random.rand() < 0.22:
                for t in range(int(self.Nswarm/10)):
                    velocity[N_cr[t],:] = -vmax + vmax*np.random.rand(self.Ndim)
                    
            #Position Updates
            x += velocity
            
            lb_violation_mask = x < lb[np.newaxis,:]
            ub_violation_mask = x > ub[np.newaxis,:]
            x_lb = lb[np.newaxis,:] + (lb[np.newaxis,:] - x)
            x_ub = ub[np.newaxis,:] - (x - ub[np.newaxis,:])
            x[lb_violation_mask] = x_lb[lb_violation_mask]
            x[ub_violation_mask] = x_ub[ub_violation_mask]
            
            self.n_iter += 1

    def run(self, n_seed, output_filename, coeff_cognitive=1.49, coeff_social=1.49, coeff_inertia=0.9, load_data=False):
        if comm.rank == 0:
            print('### Particle Swarm Optimization (seed = ' + str(n_seed) + ')\n', flush=True)
    
        np.random.seed(n_seed)
        self.output_filename = output_filename

        lb = -np.ones(self.Ndim)
        ub = np.ones(self.Ndim)

        if load_data:
            data_file1 = output_filename + "_AF_PSO_results.npz"
            data_file2 = output_filename + "_AF_PSO_density_hist.npz"
                
            with np.load(data_file1) as data:
                self.n_iter = data['n_iter']
                self.gbest_loss_hist = data['gbest_loss_hist'][:self.n_iter]

                coeff_inertia = data['coeff_inertia']
                pbest_loss = data['pbest_loss']
                
            with np.load(data_file2) as data:
                self.gbest_x_hist = data['gbest_x_hist'][:self.n_iter,:]
                self.gbest_x_binary_hist = data['gbest_x_binary_hist'][:self.n_iter,:]
                
                velocity = data['velocity']
                x = data['x']
                pbest_x = data['pbest_x']
                pbest_x_bin = data['pbest_x_bin']

        else:
            self.gbest_loss_hist = np.zeros(0)
            self.gbest_x_hist = None
            self.gbest_x_binary_hist = None
            
            velocity = None
            x = None
            pbest_loss = None
            pbest_x = None
            pbest_x_bin = None
            
            self.n_iter = 0
    
        if comm.rank == 0:
            print('    |  Iteration   |  PBest Avg.  | PBest Stdev. |  GBest Cost  | t_rem(hr) |', flush=True)
    
        self.PSO(lb, ub, coeff_inertia=coeff_inertia,
                 velocity=velocity, x=x, pbest_loss=pbest_loss, pbest_x=pbest_x, pbest_x_bin=pbest_x_bin)
        
        self.save_data()
    
    def save_data(self, coeff_inertia=0, velocity=0, x=0, pbest_loss=0, pbest_x=0, pbest_x_bin=0):
        if comm.rank == 0:
            np.savez(self.output_filename + "_AF_PSO_results",
                     gbest_loss_hist=self.gbest_loss_hist,
                     n_iter=self.n_iter,
                     coeff_inertia=coeff_inertia,
                     pbest_loss=pbest_loss)
                     
            np.savez(self.output_filename + "_AF_PSO_density_hist",
                     gbest_x_hist=self.gbest_x_hist,
                     gbest_x_binary_hist=self.gbest_x_binary_hist,
                     velocity=velocity,
                     x=x,
                     pbest_x=pbest_x,
                     pbest_x_bin=pbest_x_bin)


import torch
import torch.nn.functional as F
import numpy as np
import time

def conic_filter(mfs):
    X, Y = torch.meshgrid(torch.arange(-(mfs - 1) / 2, (mfs + 1) / 2), torch.arange(-(mfs - 1) / 2, (mfs + 1) / 2), indexing='ij')
    h = mfs / 2 - torch.sqrt(X**2 + Y**2) / (mfs / 2)
    h /= torch.sum(h)
    h[(X**2 + Y**2) > (mfs / 2)**2] = 0

    return h

def symmetrize_tensor(x, Nx, Ny, symmetry, device):
    # Detect batch mode
    batch_processing = x.ndim == 2
    if not batch_processing:
        # Standard case: x is (Ndim,)
        # Reshape to (1, Ndim) for unified processing, then squeeze at end
        x = x.unsqueeze(0)
    
    # Now x is (Batch, Ndim)
    Nbatch = x.shape[0]

    if symmetry == 0:
        x_sym = x.reshape(Nbatch, Nx, Ny)
    
    elif symmetry == 1:
        x_half = x.reshape(Nbatch, int(np.floor(Nx/2 + 0.5)), Ny)
        
        if Nx % 2 == 0:
            x_sym = torch.cat([x_half, torch.flip(x_half, [1])], dim=1)
        else:
            x_sym = torch.cat([x_half, torch.flip(x_half[:,:-1,:], [1])], dim=1)
        
    elif symmetry == 2:
        x_quart = x.reshape(Nbatch, int(np.floor(Nx/2 + 0.5)), int(np.floor(Ny/2 + 0.5)))
        
        if Ny % 2 == 0:
            top = torch.cat([x_quart, torch.flip(x_quart, [2])], dim=2)
        else:
            top = torch.cat([x_quart, torch.flip(x_quart[:,:,:-1], [2])], dim=2)
            
        if Nx % 2 == 0:
            x_sym = torch.cat([top, torch.flip(top, [1])], dim=1)
        else:
            x_sym = torch.cat([top, torch.flip(top[:,:-1,:], [1])], dim=1)
        
    elif symmetry == 4:
        assert Nx == Ny

        # Flatten just the spatial part for logic below if needed, but here we process batch
        # x is (B, Ndim)
        
        if Nx % 2 == 0:
            dim_half = int(Nx/2)
            x_quart = torch.zeros((Nbatch, dim_half, dim_half), device=device, dtype=x.dtype)
            rows, cols = torch.triu_indices(dim_half, dim_half, device=device)

            x_quart[:,rows,cols] = x
            
            # Symmetrize the quarter matrix
            # M = M + M.T - diag
            # For batch: M.transpose(1,2)
            x_quart = x_quart + x_quart.transpose(1, 2) - torch.diag_embed(torch.diagonal(x_quart, dim1=-2, dim2=-1))
            
            top_left = x_quart
            bottom_left = torch.rot90(x_quart, 1, [1, 2])
            bottom_right = torch.rot90(x_quart, 2, [1, 2])
            top_right = torch.rot90(x_quart, 3, [1, 2])
            
            top = torch.cat([top_left, top_right], dim=2)
            bottom = torch.cat([bottom_left, bottom_right], dim=2)
            x_sym = torch.cat([top, bottom], dim=1)

        else:
            dim_half = int((Nx + 1) / 2)
            x_quart = torch.zeros((Nbatch, dim_half, dim_half), device=device, dtype=x.dtype)
            rows, cols = torch.triu_indices(dim_half, dim_half, device=device)
            
            x_quart[:,rows,cols] = x
            x_quart = x_quart + x_quart.transpose(1, 2) - torch.diag_embed(torch.diagonal(x_quart, dim1=-2, dim2=-1))
            
            top_left = x_quart
            top_right = torch.rot90(x_quart, 3, [1, 2])[:,:,1:]
            bottom_left = torch.rot90(x_quart, 1, [1, 2])[:,1:,:]
            bottom_right = torch.rot90(x_quart, 2, [1, 2])[:,1:,1:]
            
            top = torch.cat([top_left, top_right], dim=2)
            bottom = torch.cat([bottom_left, bottom_right], dim=2)
            x_sym = torch.cat([top, bottom], dim=1)

    if not batch_processing:
        x_sym = x_sym.squeeze(0)
        
    return x_sym

def two_phase_projection(
        x,
        Nx,
        Ny,
        symmetry,
        periodic,
        padding,
        mfs,
        device,
        beta_proj=8,
        alpha_s=0.002,
        alpha_v=0.002,
    ):
    
    # Detect batch mode
    # x is either (Nx, Ny) or (B, Nx, Ny)
    batch_processing = x.ndim == 3
    
    beta_proj = torch.tensor(beta_proj, device=device, dtype=torch.float64)
    alpha_s = torch.tensor(alpha_s, device=device, dtype=torch.float64)
    alpha_v = torch.tensor(alpha_v, device=device, dtype=torch.float64)

    n_s = -torch.log(alpha_s)
    n_v = -torch.log(alpha_v)
    
    # Broadcast scalar/tensor ops works automatically for extra batch dim
    w_s = (1 + alpha_s) / (1 + alpha_s * torch.exp(2 * n_s * (1 - x)))
    w_v = -(1 + alpha_v) / (1 + alpha_v * torch.exp(2 * n_v * (1 + x)))
    
    kernel = conic_filter(mfs).to(dtype=torch.float64)
    padding_amount = kernel.shape[0] // 2
    
    if periodic:
        if batch_processing:
            # (B, 1, Nx, Ny)
            w_s_input = w_s.unsqueeze(1)
            w_v_input = w_v.unsqueeze(1)
        else:
            # (1, 1, Nx, Ny)
            w_s_input = w_s.unsqueeze(0).unsqueeze(0)
            w_v_input = w_v.unsqueeze(0).unsqueeze(0)

        w_s_padded = F.pad(
            w_s_input,
            (padding_amount, padding_amount, padding_amount, padding_amount),
            mode='circular',
        )
        x_s_filter = F.conv2d(
            w_s_padded,
            kernel.unsqueeze(0).unsqueeze(0),
        ).squeeze()
        
        w_v_padded = F.pad(
            w_v_input,
            (padding_amount, padding_amount, padding_amount, padding_amount),
            mode='circular',
        )
        x_v_filter = F.conv2d(
            w_v_padded,
            kernel.unsqueeze(0).unsqueeze(0),
        ).squeeze()

    else:
        # Transform the padding field to w-domain
        val_s = (1 + alpha_s) / (1 + alpha_s * torch.exp(2 * n_s * (1 - padding)))
        val_v = -(1 + alpha_v) / (1 + alpha_v * torch.exp(2 * n_v * (1 + padding)))
        
        # Required padded size
        target_h, target_w = Nx + 2 * padding_amount, Ny + 2 * padding_amount
        
        # Center crop the background if it's larger than needed
        src_h, src_w = padding.shape[-2:]
        start_h = (src_h - target_h) // 2
        start_w = (src_w - target_w) // 2
        
        # Extract background canvas
        # Padding tensor is likely (H_pad, W_pad) or maybe (1, H_pad, W_pad)
        # We need to broadcast it to batch size if in batch mode
        
        # Shared padding for all items in batch
        bg_s = val_s[start_h:start_h + target_h, start_w:start_w + target_w]
        bg_v = val_v[start_h:start_h + target_h, start_w:start_w + target_w]
        
        if batch_processing:
            Nbatch = x.shape[0]
            # Repeat for batch: (B, H, W)
            w_s_padded = bg_s.unsqueeze(0).repeat(Nbatch, 1, 1).clone()
            w_v_padded = bg_v.unsqueeze(0).repeat(Nbatch, 1, 1).clone()
        else:
            w_s_padded = bg_s.clone()
            w_v_padded = bg_v.clone()
            
        # Insert the active region w_s/w_v into the center
        # Using in-place assignment which is tracked by autograd
        if batch_processing:
            w_s_padded[:, padding_amount:padding_amount + Nx, padding_amount:padding_amount + Ny] = w_s
            w_v_padded[:, padding_amount:padding_amount + Nx, padding_amount:padding_amount + Ny] = w_v
            
            # Conv2d expects (B, C, H, W) -> (B, 1, H, W)
            w_s_in = w_s_padded.unsqueeze(1)
            w_v_in = w_v_padded.unsqueeze(1)
        else:
            w_s_padded[padding_amount:padding_amount + Nx, padding_amount:padding_amount + Ny] = w_s
            w_v_padded[padding_amount:padding_amount + Nx, padding_amount:padding_amount + Ny] = w_v
            
            # (1, 1, H, W)
            w_s_in = w_s_padded.unsqueeze(0).unsqueeze(0)
            w_v_in = w_v_padded.unsqueeze(0).unsqueeze(0)

        x_s_filter = F.conv2d(
            w_s_in,
            kernel.unsqueeze(0).unsqueeze(0),
        ).squeeze()

        x_v_filter = F.conv2d(
            w_v_in,
            kernel.unsqueeze(0).unsqueeze(0),
        ).squeeze()
    
    x_s_proj = 1 - torch.exp(-beta_proj * x_s_filter) + x_s_filter * torch.exp(-beta_proj)
    x_v_proj = -1 + torch.exp(beta_proj * x_v_filter) + x_v_filter * torch.exp(-beta_proj)

    x_proj = (x_s_proj + x_v_proj + 1) / 2

    return x_proj

class conditional_generator:
    def __init__(
        self,
        Nx,
		Ny,
		symmetry,
		periodic,
		padding,
        mfs,
        maxiter=100,
        cuda_ind=0,
    ):
        self.device = 'cuda:' + str(cuda_ind) if torch.cuda.is_available() else 'cpu'

        self.Nx = Nx
        self.Ny = Ny
        self.symmetry = symmetry
        self.periodic = periodic
        if padding is not None:
            self.padding = torch.tensor(padding, device=self.device, dtype=torch.float64)
        else:
            self.padding = padding
        self.mfs = mfs
        self.maxiter = maxiter

        # Get Number of Independent Parameters
        if symmetry == 0:
            self.Ndim = Nx*Ny
           
        elif symmetry == 1:
            self.Ndim = int(np.floor(Nx/2 + 0.5)*Ny)
        
        elif symmetry == 2:
            self.Ndim = int(np.floor(Nx/2 + 0.5)*np.floor(Ny/2 + 0.5))
        
        elif symmetry == 4:
            self.Ndim = int(np.floor(Nx/2 + 0.5)*(np.floor(Nx/2 + 0.5) + 1)/2)

    def constraint_violation(self, x_proj, epsilon=0.15):
        g = 1 - torch.abs(2 * x_proj - 1)
        c_viol = (1 / (self.Nx * self.Ny)) * torch.sum(g)

        return c_viol

    def loss(self, x, tau=0.5):
        x_proj = two_phase_projection(
            x,
            self.Nx,
            self.Ny,
            self.symmetry,
            self.periodic,
            self.padding,
            self.mfs,
            self.device,
        )
        c_viol = self.constraint_violation(x_proj)

        loss = -(1 / (self.Nx * self.Ny)) * torch.sum(self.x_reward * x_proj) + tau * c_viol

        return loss
    
    def ADAM(
        self,
        x_reward,
        beta_ADAM1=0.9,
        beta_ADAM2=0.999,
        eta_ADAM=0.1,
    ):
        # Functional approach: x_reward is the parameter we differentiate wrt
        # x is the hidden variable initiated from x_reward.
        batch_processing = x_reward.ndim == 2
        
        # Symmetrize reward (this will track gradients from x_reward)
        self.x_reward = symmetrize_tensor(x_reward, self.Nx, self.Ny, self.symmetry, self.device)
        
        if batch_processing:
            norm = torch.linalg.vector_norm(self.x_reward, dim=1, keepdim=True)
        else:
            norm = torch.linalg.vector_norm(self.x_reward)
        self.x_reward = self.x_reward / norm
        
        # Initialize x from x_reward (detached to start optimization)
        # Note: We do NOT use requires_grad_(True) because torch.func.grad handles it.
        x = self.x_reward.clone().detach()

        # Functional gradient function
        grad_fn = torch.func.grad(self.loss)
            
        jac_mean = torch.zeros_like(x)
        jac_var = torch.zeros_like(x)
        adam_iter = 0
            
        while True:
            adam_iter += 1
            
            # Loss
            loss_val = self.loss(x)
            #print(loss_val)
            
            # Compute Gradient using functional API (compatible with jacrev)
            # This computes d(loss)/dx at x
            jac = grad_fn(x)

            if adam_iter > self.maxiter:
                break
                
            # Update Average Gradients (Functional updates)
            jac_mean = beta_ADAM1*jac_mean + (1 - beta_ADAM1)*jac
            jac_var = beta_ADAM2*jac_var + (1 - beta_ADAM2)*jac**2
            
            # Unbias Average Gradients
            jac_mean_unbiased = jac_mean/(1 - beta_ADAM1**adam_iter)
            jac_var_unbiased = jac_var/(1 - beta_ADAM2**adam_iter)
            
            # Update Variables (Functional update)
            x = x - eta_ADAM*jac_mean_unbiased/(torch.sqrt(jac_var_unbiased) + 1e-8)
    
        return x
    
    def generate_near_binary(self, x_reward):
        x_opt = self.ADAM(x_reward)
        x_near_binary = two_phase_projection(
            x_opt,
            self.Nx,
            self.Ny,
            self.symmetry,
            self.periodic,
            self.padding,
            self.mfs,
            self.device,
        ).reshape(-1)

        return x_near_binary, x_near_binary.detach()

    def generate_near_binary_no_grad(self, x_reward):
        # Run optimization without tracking gradients w.r.t x_reward to save memory
        # We detach x_reward so the optimization graph is not built
        x_reward_detached = x_reward.detach()
        x_opt = self.ADAM(x_reward_detached)
        x_near_binary = two_phase_projection(
            x_opt,
            self.Nx,
            self.Ny,
            self.symmetry,
            self.periodic,
            self.padding,
            self.mfs,
            self.device,
        ).reshape(x_reward.shape[0], -1)

        # Return detached result (though it should already be detached effectively)
        return x_near_binary.detach()


import numpy as np
from scipy.ndimage import gaussian_filter
from scipy.interpolate import interpn
from itertools import product
import gegd.parameter_processing.symmetry_operations as symOp
import gegd.parameter_processing.density_transforms as dtf
import time

from mpi4py import MPI
comm = MPI.COMM_WORLD

class optimizer:
    def __init__(self,
                 Nx,
                 Ny,
                 symmetry,
                 periodic,
                 padding,
                 maxiter,
                 t_low_fidelity,
                 t_high_fidelity,
                 t_iteration,
                 high_fidelity_setting,
                 low_fidelity_setting,
                 brush_size,
                 sigma_ensemble_max=1.0,
                 upsample_ratio=1,
                 beta_proj=8,
                 brush_shape='circle',
                 covariance_type='constant',
                 coeff_exp=5,
                 cost_threshold=0,
                 cost_obj=None,
                 Nthreads=1,
                 ):
        
        self.Nx = Nx
        self.Ny = Ny
        self.symmetry = symmetry
        self.periodic = periodic
        self.padding = padding
        self.maxiter = maxiter
        self.t_low_fidelity = t_low_fidelity
        self.t_high_fidelity = t_high_fidelity
        self.t_iteration = t_iteration
        self.beta_proj = beta_proj
        self.beta_proj_sigma = beta_proj
        self.brush_size = brush_size
        self.sigma_ensemble_max = sigma_ensemble_max
        self.upsample_ratio = upsample_ratio
        self.brush_shape = brush_shape
        self.sigma_filter = brush_size/2/np.sqrt(2)
        self.sigma_RBF = brush_size/2/np.sqrt(2)
        self.covariance_type = covariance_type
        self.high_fidelity_setting = high_fidelity_setting
        self.low_fidelity_setting = low_fidelity_setting
        self.coeff_exp = coeff_exp
        self.cost_threshold = cost_threshold
        self.cost_obj = cost_obj
        self.Nthreads = Nthreads
        
        # Get Number of Independent Parameters
        if symmetry == 0:
            self.Ndim = Nx*Ny
           
        elif symmetry == 1:
            self.Ndim = int(np.floor(Nx/2 + 0.5)*Ny)
        
        elif symmetry == 2:
            self.Ndim = int(np.floor(Nx/2 + 0.5)*np.floor(Ny/2 + 0.5))
        
        elif symmetry == 4:
            self.Ndim = int(np.floor(Nx/2 + 0.5)*(np.floor(Nx/2 + 0.5) + 1)/2)
        
        if self.covariance_type == 'constant':
            self.Nsigma = 1
            
        elif self.covariance_type == 'diagonal':
            self.Nsigma = self.Ndim
            
        elif self.covariance_type == 'gaussian_constant':
            self.Nsigma = 1
            self.construct_gaussian_covariance()
            
        elif self.covariance_type == 'gaussian_diagonal':
            self.Nsigma = self.Ndim
            self.construct_gaussian_covariance()
    
    def construct_gaussian_covariance(self, max_condition_number=1e4):
        self.cov_g = np.zeros((self.Ndim, self.Ndim))
        
        if self.symmetry == 0:
            delta = np.zeros((self.Nx, self.Ny))
            delta[0,0] = 1
            if self.periodic:
                kernel = gaussian_filter(delta, self.sigma_RBF, mode='wrap')
            else:
                kernel = gaussian_filter(delta, self.sigma_RBF, mode='constant')
        
            for i in range(self.Nx):
                for j in range(self.Ny):
                    self.cov_g[self.Ny*i+j,:] = np.roll(kernel, (i, j), axis=(0,1)).reshape(-1)
        
        elif self.symmetry == 1:
            DOF_x = int(np.floor(self.Nx/2 + 0.5))
        
            for i in range(DOF_x):
                for j in range(self.Ny):
                    delta = np.zeros((self.Nx, self.Ny))
                    if self.Nx % 2 == 0:
                        c = 1
                    
                    elif self.Nx % 2 == 1:
                        if i == DOF_x - 1:
                            c = 2
                        else:
                            c = 1
                    
                    delta[i,j] = c
                    delta = symOp.symmetrize(delta[:DOF_x,:], self.symmetry, self.Nx, self.Ny)
                    if self.periodic:
                        kernel = gaussian_filter(delta, self.sigma_RBF, mode='wrap')
                    else:
                        kernel = gaussian_filter(delta, self.sigma_RBF, mode='constant')
                    
                    self.cov_g[i*self.Ny+j,:] = symOp.desymmetrize(kernel, self.symmetry, self.Nx, self.Ny)
        
        elif self.symmetry == 2:
            DOF_x = int(np.floor(self.Nx/2 + 0.5))
            DOF_y = int(np.floor(self.Ny/2 + 0.5))
            
            for i in range(DOF_x):
                for j in range(DOF_y):
                    delta = np.zeros((self.Nx, self.Ny))
                    c = 1
                    if self.Nx % 2 == 1:
                        if i == DOF_x - 1:
                            c *= 2
                    
                    if self.Ny % 2 == 1:
                        if j == DOF_y - 1:
                            c *= 2
                    
                    delta[i,j] = c
                    delta = symOp.symmetrize(delta[:DOF_x,:DOF_y], self.symmetry, self.Nx, self.Ny)
                    if self.periodic:
                        kernel = gaussian_filter(delta, self.sigma_RBF, mode='wrap')
                    else:
                        kernel = gaussian_filter(delta, self.sigma_RBF, mode='constant')
                    
                    self.cov_g[i*DOF_y+j,:] = symOp.desymmetrize(kernel, self.symmetry, self.Nx, self.Ny)
        
        elif self.symmetry == 4:
            triu_ind = np.triu_indices(int(np.floor(self.Nx/2 + 0.5)))
            
            for i in range(self.Ndim):
                delta = np.zeros((self.Nx, self.Ny))
                if self.Nx % 2 == 0:
                    if triu_ind[0][i] == triu_ind[1][i]:
                        c = 2
                    else:
                        c = 1
                
                elif self.Nx % 2 == 1:
                    if i == self.Ndim - 1:
                        c = 8
                    elif triu_ind[1][i] == int((self.Nx - 1)/2):
                        c = 2
                    elif triu_ind[0][i] == triu_ind[1][i]:
                        c = 2
                    else:
                        c = 1
                
                delta[triu_ind[0][i],triu_ind[1][i]] = c
                delta = symOp.symmetrize(delta[triu_ind], self.symmetry, self.Nx, self.Ny)
                if self.periodic:
                    kernel = gaussian_filter(delta, self.sigma_RBF, mode='wrap')
                else:
                    kernel = gaussian_filter(delta, self.sigma_RBF, mode='constant')
                
                self.cov_g[i,:] = symOp.desymmetrize(kernel, self.symmetry, self.Nx, self.Ny)
        
        self.cov_g /= np.max(self.cov_g)
        
        # Determine minimum eps for inversion
        eigvals = np.linalg.eigvalsh(self.cov_g)
        lambda_max = np.max(eigvals)
        lambda_min = np.min(eigvals)
        eps = (lambda_max - max_condition_number*lambda_min)/(max_condition_number - 1)
        eps = max(eps, 0)
        
        self.cov_g = self.cov_g + eps*np.identity(self.Ndim)
        self.cov_g_inv = np.linalg.inv(self.cov_g)
    
    def construct_cov_matrix(self, sigma):
        if self.covariance_type == 'constant':
            cov = sigma**2*np.identity(self.Ndim)
            cov_inv = (1/sigma**2)*np.identity(self.Ndim)
        
        elif self.covariance_type == 'diagonal':
            cov = np.diag(sigma**2)
            cov_inv = np.diag(1/sigma**2)
        
        elif self.covariance_type == 'gaussian_constant':
            cov = sigma**2*self.cov_g
            cov_inv = (1/sigma**2)*self.cov_g_inv
        
        elif self.covariance_type == 'gaussian_diagonal':
            cov = np.diag(sigma) @ self.cov_g @ np.diag(sigma)
            cov_inv = np.diag(1/sigma) @ self.cov_g_inv @ np.diag(1/sigma)
        
        return cov, cov_inv
    
    def mu_derivative(self, dx, cov_inv, p):
        
        return (cov_inv @ dx)*p
    
    def sigma_derivative(self, dx, sigma, cov_inv, p):
        if self.covariance_type == 'constant':
            grad = ((dx.reshape(1,-1) @ dx.reshape(-1,1))/sigma**3 - self.Ndim/sigma)*p
        
        elif self.covariance_type == 'diagonal':
            grad = np.diag(cov_inv @ dx.reshape(-1,1) @ dx.reshape(1,-1) @ np.sqrt(cov_inv) - np.sqrt(cov_inv))*p
            
        elif self.covariance_type == 'gaussian_constant':
            grad = (1/sigma)*(dx.reshape(1,-1) @ cov_inv @ dx.reshape(-1,1) - self.Ndim)*p
        
        elif self.covariance_type == 'gaussian_diagonal':
            z = cov_inv @ dx
            Bz = (self.cov_g*sigma[np.newaxis,:]) @ z
            A_diag = z*Bz
        
            grad = (A_diag - 1/sigma)*p
        
        return grad
    
    def ensemble_jacobian(self, x0, dx, sigma_ensemble, test_function):
        x0 = x0[:self.Ndim]
        cov, cov_inv = self.construct_cov_matrix(sigma_ensemble)
        
        # Get Brush Binarized Densities ------------------------------------------------------------
        x_sample = dtf.binarize(x0, self.symmetry, self.periodic, self.Nx, self.Ny, self.brush_size, self.brush_shape, self.beta_proj, self.sigma_filter, dx=dx,
                                upsample_ratio=self.upsample_ratio, padding=self.padding, Nthreads=self.Nthreads)
        
        # Sample Modified Cost Function --------------------------------------------------------------
        self.cost_obj.set_accuracy(self.high_fidelity_setting)
        
        f = np.zeros(self.Nensemble)
        f_logDeriv = np.zeros((self.Nensemble, self.Ndim+self.Nsigma))
        for n in range(self.Nensemble):
            f_temp = self.cost_obj.get_cost(x_sample[n,:], False)
            f_shifted = (f_temp - self.cost_threshold)/(self.cost_threshold + 1)
            f[n] = -np.exp(-self.coeff_exp*f_shifted)
            f_logDeriv[n,:self.Ndim] = self.mu_derivative(dx[n,:], cov_inv, f[n])
            f_logDeriv[n,self.Ndim:] = self.sigma_derivative(dx[n,:], sigma_ensemble, cov_inv, f[n])

        if test_function:
            x_fp0, x_b0 = dtf.binarize(x0.reshape(1,-1), self.symmetry, self.periodic, self.Nx, self.Ny, self.brush_size, self.brush_shape, self.beta_proj, self.sigma_filter,
                                       upsample_ratio=self.upsample_ratio, output_details=True, padding=self.padding, Nthreads=self.Nthreads)
            f0, jac_b_STE_sym = self.cost_obj.get_cost(x_b0, True)
            jac_b_STE_sym = jac_b_STE_sym.reshape(self.Nx, self.Ny)
        else:
            x_fp0 = np.zeros_like(x0)
            x_b0 = np.zeros_like(x0)
            f0 = 0
        
        # Get Best Cost -----------------------------------------------------------------------------------
        f_best = (-np.log(-np.min(f))/self.coeff_exp)*(self.cost_threshold + 1) + self.cost_threshold
        x_best = x_sample[np.argmin(f),:].copy()
        
        # Sample Control Variate Function ----------------------------------------------------------------
        self.cost_obj.set_accuracy(self.low_fidelity_setting)
        
        f_ctrl_all = np.zeros(self.r_CV*self.Nensemble)
        f_ctrl_logDeriv_all = np.zeros((self.r_CV*self.Nensemble, self.Ndim+self.Nsigma))
        for n in range(self.r_CV*self.Nensemble):
            f_temp = self.cost_obj.get_cost(x_sample[n,:], False)
            f_shifted = (f_temp - self.cost_threshold)/(self.cost_threshold + 1)
            f_ctrl_all[n] = -np.exp(-self.coeff_exp*f_shifted)
            f_ctrl_logDeriv_all[n,:self.Ndim] = self.mu_derivative(dx[n,:], cov_inv, f_ctrl_all[n])
            f_ctrl_logDeriv_all[n,self.Ndim:] = self.sigma_derivative(dx[n,:], sigma_ensemble, cov_inv, f_ctrl_all[n])
        
        # Expectation of the Control Variate Function -----------------------------------------------------
        ctrlVarExp_f = np.mean(f_ctrl_all)
        ctrlVarExp_mu = np.mean(f_ctrl_logDeriv_all[:,:self.Ndim], axis=0)
        ctrlVarExp_sigma = np.mean(f_ctrl_logDeriv_all[:,self.Ndim:], axis=0)

        # Control Variate Coefficient (expectation) --------------------------------------------------------------------
        f_ctrl = f_ctrl_all[:self.Nensemble]
        mean_f = np.mean(f)
        mean_f_ctrl = np.mean(f_ctrl)
        
        cov_f_f_ctrl = (1/(self.Nensemble - 1))*np.sum((f - mean_f)*(f_ctrl - mean_f_ctrl))
        var_f = (1/(self.Nensemble - 1))*np.sum((f - mean_f)**2)
        var_ctrl = (1/(self.Nensemble - 1))*np.sum((f_ctrl - mean_f_ctrl)**2)

        ctrlVarCoeff_f = cov_f_f_ctrl/var_ctrl
        corr_f = cov_f_f_ctrl/np.sqrt(var_f*var_ctrl)
        
        # Control Variate Coefficient (expectation gradient wrt mu) --------------------------------------------------------------------
        f_mu = f_logDeriv[:,:self.Ndim]
        f_ctrl_mu = f_ctrl_logDeriv_all[:self.Nensemble,:self.Ndim]
        mean_f_mu = np.mean(f_mu, axis=0)
        mean_f_ctrl_mu = np.mean(f_ctrl_mu, axis=0)
        
        cov_f_f_ctrl = (1/(self.Nensemble - 1))*np.sum((f_mu - mean_f_mu)*(f_ctrl_mu - mean_f_ctrl_mu), axis=0)
        var_f = (1/(self.Nensemble - 1))*np.sum((f_mu - mean_f_mu)**2, axis=0)
        var_ctrl = (1/(self.Nensemble - 1))*np.sum((f_ctrl_mu - mean_f_ctrl_mu)**2, axis=0)

        weights = np.ones(self.Ndim) #1/np.hstack((sigma_ensemble, sigma_ensemble))**2
        ctrlVarCoeff_mu = np.average(cov_f_f_ctrl, weights=weights)/np.average(var_ctrl, weights=weights)
        corr_mu = cov_f_f_ctrl/np.sqrt(var_f*var_ctrl)
        
        # Control Variate Coefficient (expectation gradient wrt sigma) --------------------------------------------------------------------
        f_sigma = f_logDeriv[:,self.Ndim:]
        f_ctrl_sigma = f_ctrl_logDeriv_all[:self.Nensemble,self.Ndim:]
        mean_f_sigma = np.mean(f_sigma, axis=0)
        mean_f_ctrl_sigma = np.mean(f_ctrl_sigma, axis=0)
        
        cov_f_f_ctrl = (1/(self.Nensemble - 1))*np.sum((f_sigma - mean_f_sigma)*(f_ctrl_sigma - mean_f_ctrl_sigma), axis=0)
        var_f = (1/(self.Nensemble - 1))*np.sum((f_sigma - mean_f_sigma)**2, axis=0)
        var_ctrl = (1/(self.Nensemble - 1))*np.sum((f_ctrl_sigma - mean_f_ctrl_sigma)**2, axis=0)

        weights = np.ones(self.Nsigma) #1/np.hstack((sigma_ensemble, sigma_ensemble))**2
        ctrlVarCoeff_sigma = np.average(cov_f_f_ctrl, weights=weights)/np.average(var_ctrl, weights=weights)
        corr_sigma = cov_f_f_ctrl/np.sqrt(var_f*var_ctrl)
        
        # Control Variate Estimation of the expectation --------------------------------------------------------------
        if self.r_CV == 1:
            f_est = f.copy()
        else:
            f_est = f - ctrlVarCoeff_f*(f_ctrl - ctrlVarExp_f)
        
        # Control Variate Estimation of the gradient --------------------------------------------------------------
        jac_fp_ensemble = f_logDeriv.copy()
        if self.r_CV > 1:
            jac_fp_ensemble[:,:self.Ndim] -= ctrlVarCoeff_mu*(f_ctrl_mu - ctrlVarExp_mu)
            jac_fp_ensemble[:,self.Ndim:] -= ctrlVarCoeff_sigma*(f_ctrl_sigma - ctrlVarExp_sigma)
        jac_fp_ensemble = np.mean(jac_fp_ensemble, axis=0)

        jac_mu_fp_ensemble_sym = symOp.symmetrize_jacobian(jac_fp_ensemble[:self.Ndim], self.symmetry, self.Nx, self.Ny)
        if self.Nsigma > 1:
            jac_sigma_fp_ensemble_sym = symOp.symmetrize_jacobian(jac_fp_ensemble[self.Ndim:], self.symmetry, self.Nx, self.Ny)
        
        if test_function:
            jac_latent_STE = dtf.backprop_filter_and_project(jac_b_STE_sym, x0, self.symmetry, self.periodic, self.Nx, self.Ny, self.brush_size, self.sigma_filter, self.beta_proj, padding=self.padding)
        
        jac_latent_ensemble = jac_fp_ensemble.copy()
        jac_latent_ensemble[:self.Ndim] = dtf.backprop_filter_and_project(jac_mu_fp_ensemble_sym, x0, self.symmetry, self.periodic, self.Nx, self.Ny, self.brush_size, self.sigma_filter, self.beta_proj, padding=self.padding)
        if self.Nsigma > 1:
            jac_latent_ensemble[self.Ndim:] = dtf.backprop_filter_and_project(jac_sigma_fp_ensemble_sym, x0, self.symmetry, self.periodic, self.Nx, self.Ny, self.brush_size, self.sigma_filter, self.beta_proj_sigma, padding=self.padding)
        
        if test_function:
            cost_all = (f0, np.mean(f_est))
            density_all = (x_fp0, x_b0)
            grad_STE = (jac_b_STE_sym, jac_latent_STE)
            grad_ensemble = (jac_mu_fp_ensemble_sym, jac_latent_ensemble)
        
            return cost_all, density_all, grad_STE, grad_ensemble
        else:
            return f0, x_b0, np.mean(f), np.std(f), np.mean(f_ctrl_all), np.std(f_ctrl_all), jac_latent_ensemble, ctrlVarCoeff_mu, ctrlVarCoeff_sigma, corr_f, corr_mu, corr_sigma, f_best, x_best
    
    def ADAM(self,
             x_bounded,
             lb,
             ub,
             beta_ADAM1,
             beta_ADAM2,
             eta_ADAM,
             jac_mean=None,
             jac_var=None,
             adam_iter=None,
             x_bounded_norm_ref=None,
             ):
    
        # Dummy Variables
        x = -np.log((ub - lb)/(x_bounded - lb) - 1)
        x[ub==lb] = ub[ub==lb]
        
        if jac_mean is None:
            jac_mean = np.zeros_like(x)
        if jac_var is None:
            jac_var = np.zeros_like(x)
        if adam_iter is None:
            adam_iter = 0
            
        while True:
            t1 = time.time()
            adam_iter += 1
            
            # Determine optimal ensemble sizes for the high and low-fidelity simulations given the target iteration time
            N = np.arange(int(np.floor(self.t_iteration/self.t_high_fidelity))) + 1
            N = N[N>=5]
            r_CV = np.floor(self.t_iteration/(self.t_low_fidelity*N) - self.t_high_fidelity/self.t_low_fidelity).astype(np.int32)
                        
            N = N[(r_CV>0)*(r_CV<=100)]
            r_CV = r_CV[(r_CV>0)*(r_CV<=100)]
            
            if self.corr_mu_hist is None:
                corr_mean = 0
            elif self.corr_mu_hist.ndim == 1:
                corr_mean = np.mean(self.corr_mu_hist)
            else:
                corr_mean = np.mean(self.corr_mu_hist[-1,:])
            
            var_reduction = np.zeros(N.size)
            var_reduction[r_CV>0] = (1 - ((r_CV[r_CV>0] - 1)/r_CV[r_CV>0])*corr_mean**2)/N[r_CV>0]
            var_reduction[r_CV<=0] = 1/N[r_CV<=0]
            ind_opt = np.nanargmin(var_reduction)
            self.Nensemble = N[ind_opt]
            self.r_CV = r_CV[ind_opt]
            
            var_reduction = (1 - ((self.r_CV - 1)/self.r_CV)*corr_mean**2)/self.Nensemble
            
            x_bounded = lb + (ub - lb)/(1 + np.exp(-x))
            
            sigma_ensemble = x_bounded[self.Ndim:]
            if self.Nsigma == 1:
                sigma_fp = sigma_ensemble.copy()
            else:
                sigma_scaled = 2*(sigma_ensemble - lb[self.Ndim:])/(ub[self.Ndim:] - lb[self.Ndim:]) - 1
                sigma_fp_scaled = dtf.filter_and_project(sigma_scaled, self.symmetry, self.periodic, self.Nx, self.Ny, self.brush_size, self.sigma_filter, self.beta_proj_sigma, padding=self.padding)
                sigma_fp = (sigma_fp_scaled + 1)*(ub[self.Ndim:] - lb[self.Ndim:])/2 + lb[self.Ndim:]
            cov, cov_inv = self.construct_cov_matrix(sigma_fp)
            
            dx = np.random.multivariate_normal(np.zeros(self.Ndim), cov, size=self.r_CV*self.Nensemble)
            
            loss, x_bin0, loss_mean, loss_std, loss_ctrl_mean, loss_ctrl_std, jac, ctrlVarCoeff_mu, ctrlVarCoeff_sigma, corr_f, corr_mu, corr_sigma, f_best, x_best = self.ensemble_jacobian(x_bounded,
                                                                                                                                                                                             dx,
                                                                                                                                                                                             sigma_fp,
                                                                                                                                                                                             False)
            
            if self.Nsigma > 1:
                jac[self.Ndim:] *= 2/(ub[self.Ndim:] - lb[self.Ndim:]) # accounts for covariance rescaling
            
            jac_save = jac.copy()

            jac[ub==lb] = 0
            jac *= np.exp(-x)*(ub - lb)/(1 + np.exp(-x))**2
            
            self.N_high_fidelity_hist = np.append(self.N_high_fidelity_hist, self.Nensemble)
            self.N_low_fidelity_hist = np.append(self.N_low_fidelity_hist, self.r_CV*self.Nensemble)
            self.var_reduction_hist = np.append(self.var_reduction_hist, var_reduction)
            self.N_eff_hist = np.append(self.N_eff_hist, int(1/var_reduction))
            
            if self.x_latent_hist is None:
                self.x_latent_hist = x_bounded[:self.Ndim].copy()
            else:
                self.x_latent_hist = np.vstack((self.x_latent_hist, x_bounded[:self.Ndim]))
                
            if self.x_hist is None:
                self.x_hist = x_bin0.copy()
            else:
                self.x_hist = np.vstack((self.x_hist, x_bin0))
            
            if self.best_cost_hist.size == 0:
                new_best = True
                self.best_cost_hist = np.append(self.best_cost_hist, f_best)
            else:
                new_best = self.best_cost_hist[-1] > f_best
                self.best_cost_hist = np.append(self.best_cost_hist, np.min((self.best_cost_hist[-1], f_best)))
                
            if self.best_x_hist is None:
                self.best_x_hist = x_best.copy()
            else:
                if new_best:
                    self.best_x_hist = np.vstack((self.best_x_hist, x_best))
                else:
                    if self.best_x_hist.ndim == 1:
                        self.best_x_hist = np.vstack((self.best_x_hist, self.best_x_hist))
                    else:
                        self.best_x_hist = np.vstack((self.best_x_hist, self.best_x_hist[-1,:]))
            
            self.cost_ensemble_hist = np.append(self.cost_ensemble_hist, loss_mean)
            self.cost_ensemble_sigma_hist = np.append(self.cost_ensemble_sigma_hist, loss_std)
            self.cost_ensemble_ctrl_hist = np.append(self.cost_ensemble_ctrl_hist, loss_ctrl_mean)
            self.cost_ensemble_ctrl_sigma_hist = np.append(self.cost_ensemble_ctrl_sigma_hist, loss_ctrl_std)
            
            if self.Nsigma == 1:
                self.sigma_ensemble_hist = np.append(self.sigma_ensemble_hist, sigma_ensemble)
            else:
                if self.sigma_ensemble_hist is None:
                    self.sigma_ensemble_hist = sigma_ensemble.copy()
                else:
                    self.sigma_ensemble_hist = np.vstack((self.sigma_ensemble_hist, sigma_ensemble))
                
                if self.sigma_fp_hist is None:
                    self.sigma_fp_hist = sigma_fp.copy()
                else:
                    self.sigma_fp_hist = np.vstack((self.sigma_fp_hist, sigma_fp))
            
            self.ctrlVarCoeff_mu_hist = np.append(self.ctrlVarCoeff_mu_hist, ctrlVarCoeff_mu)
            self.ctrlVarCoeff_sigma_hist = np.append(self.ctrlVarCoeff_sigma_hist, ctrlVarCoeff_sigma)
            self.corr_f_hist = np.append(self.corr_f_hist, corr_f)
            
            if self.corr_mu_hist is None:
                self.corr_mu_hist = corr_mu.copy()
            else:
                self.corr_mu_hist = np.vstack((self.corr_mu_hist, corr_mu))
            if self.corr_sigma_hist is None:
                self.corr_sigma_hist = corr_sigma.copy()
            else:
                self.corr_sigma_hist = np.vstack((self.corr_sigma_hist, corr_sigma))
        
            t2 = time.time()
            t_rem = (t2 - t1)*(self.maxiter - self.n_iter + 1)/3600
            
            if comm.rank == 0:
                if self.Nsigma == 1:
                    print('    | %4d |  %4d |   %4d  | %7.5f |  %4d | %7.5f |  %5.2f  | %7.4f | %7.4f |  %9.2f |  %9.2f  |  %8.3f |' %(self.n_iter,
                                                                                                                                       self.Nensemble,
                                                                                                                                       self.r_CV*self.Nensemble,
                                                                                                                                       var_reduction,
                                                                                                                                       int(1/var_reduction),
                                                                                                                                       sigma_fp[0],
                                                                                                                                       ctrlVarCoeff_mu,
                                                                                                                                       np.mean(corr_f),
                                                                                                                                       np.mean(corr_mu),
                                                                                                                                       -np.log(-loss_mean)/self.coeff_exp,
                                                                                                                                       -np.log(-loss_ctrl_mean)/self.coeff_exp,
                                                                                                                                       self.best_cost_hist[-1]),
                                                                                                                                       end='', flush=True)
                
                else:
                    print('    | %4d |  %4d |   %4d  | %7.5f |  %4d | %7.5f | %7.5f |  %5.2f  | %7.4f | %7.4f |  %9.2f | %7.2f |  %9.2f  |   %7.2f  |  %8.3f |' %(self.n_iter,
                                                                                                                                                                  self.Nensemble,
                                                                                                                                                                  self.r_CV*self.Nensemble,
                                                                                                                                                                  var_reduction,
                                                                                                                                                                  int(1/var_reduction),
                                                                                                                                                                  np.min(sigma_fp),
                                                                                                                                                                  np.max(sigma_fp),
                                                                                                                                                                  ctrlVarCoeff_mu,
                                                                                                                                                                  np.mean(corr_f),
                                                                                                                                                                  np.mean(corr_mu),
                                                                                                                                                                  loss_mean,
                                                                                                                                                                  loss_std,
                                                                                                                                                                  loss_ctrl_mean,
                                                                                                                                                                  loss_ctrl_std,
                                                                                                                                                                  self.best_cost_hist[-1]),
                                                                                                                                                                  end='', flush=True)
            self.save_data(x_bounded=x_bounded,
                           jac_mean=jac_mean,
                           jac_var=jac_var,
                           adam_iter=adam_iter,
                           x_bounded_norm_ref=x_bounded_norm_ref)

            if adam_iter >= self.maxiter:
                #self.n_iter += 1
                break
                
            # Update Average Gradients
            jac_mean = beta_ADAM1*jac_mean + (1 - beta_ADAM1)*jac
            jac_var = beta_ADAM2*jac_var + (1 - beta_ADAM2)*jac**2
            
            # Unbias Average Gradients
            jac_mean_unbiased = jac_mean/(1 - beta_ADAM1**adam_iter)
            jac_var_unbiased = jac_var/(1 - beta_ADAM2**adam_iter)
            
            # Update Variables
            if self.n_iter == 0:
                x_bounded_norm_ref = 1
                x_bounded_norm = 1
            elif self.n_iter == 1:
                x_bounded_norm_ref = np.linalg.norm(x_bounded[:self.Ndim])
                x_bounded_norm = np.linalg.norm(x_bounded[:self.Ndim])
            else:
                x_bounded_norm = np.linalg.norm(x_bounded[:self.Ndim])

            #eta_sched = eta_ADAM.copy()
            eta_sched = eta_ADAM*(x_bounded_norm/x_bounded_norm_ref)**(1/3)
            if comm.rank == 0:
                print('  %6.4f  | %7.5f |   %5.2f   |' %(x_bounded_norm, eta_sched[0], t_rem), flush=True)
            x -= eta_sched*jac_mean_unbiased/(np.sqrt(jac_var_unbiased) + 1e-8)
            # eta *= 0.95
            
            self.x_bounded_norm_hist = np.append(self.x_bounded_norm_hist, x_bounded_norm)
            self.eta_sched_hist = np.append(self.eta_sched_hist, eta_sched[0])
            
            self.n_iter += 1

    def run(self, n_seed, output_filename, x0=None, eta_mu=0.01, eta_sigma=1.0, load_data=False):
        if comm.rank == 0:
            print('### Ensemble Optimization (seed = ' + str(n_seed) + ')\n', flush=True)
    
        np.random.seed(n_seed)
        self.output_filename = output_filename
        
        sigma_ensemble = self.sigma_ensemble_max/2

        lb = np.zeros(self.Ndim+self.Nsigma)
        lb[:self.Ndim] = -1
        lb[self.Ndim:] = 0.99*self.sigma_ensemble_max/2
        ub = np.zeros(self.Ndim+self.Nsigma)
        ub[:self.Ndim] = 1
        ub[self.Ndim:] = 1.01*self.sigma_ensemble_max/2
        #ub[self.Ndim:] = 10.0
        
        if load_data:
            data_file1 = output_filename + "_GEGD_results.npz"
            data_file2 = output_filename + "_GEGD_density_hist.npz"
                
            with np.load(data_file1) as data:
                self.n_iter = data['n_iter']
                
                self.N_high_fidelity_hist = data['N_high_fidelity_hist'][:self.n_iter]
                self.N_low_fidelity_hist = data['N_low_fidelity_hist'][:self.n_iter]
                self.var_reduction_hist = data['var_reduction_hist'][:self.n_iter]
                self.N_eff_hist = data['N_eff_hist'][:self.n_iter]
                self.cost_ensemble_hist = data['cost_ensemble_hist'][:self.n_iter]
                self.cost_ensemble_ctrl_hist = data['cost_ensemble_ctrl_hist'][:self.n_iter]
                self.best_cost_hist = data['best_cost_hist'][:self.n_iter]
                self.cost_ensemble_sigma_hist = data['cost_ensemble_sigma_hist'][:self.n_iter]
                self.cost_ensemble_ctrl_sigma_hist = data['cost_ensemble_ctrl_sigma_hist'][:self.n_iter]
                self.ctrlVarCoeff_mu_hist = data['ctrlVarCoeff_mu_hist'][:self.n_iter]
                self.ctrlVarCoeff_sigma_hist = data['ctrlVarCoeff_sigma_hist'][:self.n_iter]
                self.corr_f_hist = data['corr_f_hist'][:self.n_iter]
                self.x_bounded_norm_hist = data['x_bounded_norm_hist'][:self.n_iter]
                self.eta_sched_hist = data['eta_sched_hist'][:self.n_iter]
                if self.Nsigma == 1:
                    self.sigma_ensemble_hist = data['sigma_ensemble_hist'][:self.n_iter]
                adam_iter = data['adam_iter'] - 1
                x_bounded_norm_ref = data['x_bounded_norm_ref']
                
            with np.load(data_file2) as data:
                self.x_latent_hist = data['x_latent_hist'][:self.n_iter,:]
                self.x_hist = data['x_hist'][:self.n_iter,:]
                self.best_x_hist = data['best_x_hist'][:self.n_iter,:]
                if self.Nsigma > 1:
                    self.sigma_ensemble_hist = data['sigma_ensemble_hist'][:self.n_iter,:]
                    self.sigma_fp_hist = data['sigma_fp_hist'][:self.n_iter,:]
                self.corr_mu_hist = data['corr_mu_hist'][:self.n_iter,:]
                self.corr_sigma_hist = data['corr_sigma_hist'][:self.n_iter,:]
                x0 = data['x_bounded']
                jac_mean = data['jac_mean']
                jac_var = data['jac_var']
            
        else:
            self.n_iter = 0
            
            self.N_high_fidelity_hist = np.zeros(0)
            self.N_low_fidelity_hist = np.zeros(0)
            self.var_reduction_hist = np.zeros(0)
            self.N_eff_hist = np.zeros(0)
            self.x_latent_hist = None
            self.x_hist = None
            self.best_x_hist = None
            self.cost_ensemble_hist = np.zeros(0)
            self.cost_ensemble_ctrl_hist = np.zeros(0)
            self.best_cost_hist = np.zeros(0)
            self.cost_ensemble_sigma_hist = np.zeros(0)
            self.cost_ensemble_ctrl_sigma_hist = np.zeros(0)
            if self.Nsigma == 1:
                self.sigma_ensemble_hist = np.zeros(0)
            else:
                self.sigma_ensemble_hist = None
                self.sigma_fp_hist = None
            self.ctrlVarCoeff_mu_hist = np.zeros(0)
            self.ctrlVarCoeff_sigma_hist = np.zeros(0)
            self.corr_mu_hist = None
            self.corr_sigma_hist = None
            self.corr_f_hist = np.zeros(0)
            self.x_bounded_norm_hist = np.zeros(0)
            self.eta_sched_hist = np.zeros(0)
            
            if x0 is None:
                # Initial Structure
                x0 = np.hstack((np.zeros(self.Ndim), sigma_ensemble*np.ones(self.Nsigma)))
            
            jac_mean = None
            jac_var = None
            adam_iter = None
            x_bounded_norm_ref = None
        
        if comm.rank == 0:
            if self.Nsigma == 1:
                print('    | Iter | N_acc | N_inacc | var_red | N_eff |  sigma  | CVCoeff | corr(f) | corr(g) |  Cost Ens  | Cost Ens CV | Cost Best |  x norm  | ADAM_LR | t_rem(hr) |',
                      flush=True)
            else:
                print('    | Iter | N_acc | N_inacc | var_red | N_eff | sig_min | sig_max | CVCoeff | corr(f) | corr(g) |  Cost Ens  | StD Ens | Cost Ens CV | StD Ens CV | Cost Best |  x norm  | ADAM_LR | t_rem(hr) |',
                      flush=True)

        eta_ADAM = np.ones(self.Ndim+self.Nsigma)
        eta_ADAM[:self.Ndim] = eta_mu
        eta_ADAM[self.Ndim:] = eta_sigma
        self.ADAM(x0, lb, ub, 0.9, 0.999, eta_ADAM,
                  jac_mean=jac_mean, jac_var=jac_var, adam_iter=adam_iter, x_bounded_norm_ref=x_bounded_norm_ref)
        
        self.save_data()
    
    def save_data(self, x_bounded=None, jac_mean=None, jac_var=None, adam_iter=None, x_bounded_norm_ref=None):
        if comm.rank == 0:
            if x_bounded is None:
                with np.load(self.output_filename + "_GEGD_results.npz") as data:
                    adam_iter = data['adam_iter']
                    x_bounded_norm_ref = data['x_bounded_norm_ref']
                
                with np.load(self.output_filename + "_GEGD_density_hist.npz") as data:
                    x_bounded = data['x_bounded']
                    jac_mean = data['jac_mean']
                    jac_var = data['jac_var']
        
            # Customize below
            if self.Nsigma == 1:
                np.savez(self.output_filename + "_GEGD_results",
                         N_high_fidelity_hist=self.N_high_fidelity_hist,
                         N_low_fidelity_hist=self.N_low_fidelity_hist,
                         var_reduction_hist=self.var_reduction_hist,
                         N_eff_hist=self.N_eff_hist,
                         cost_ensemble_hist=self.cost_ensemble_hist,
                         cost_ensemble_sigma_hist=self.cost_ensemble_sigma_hist,
                         cost_ensemble_ctrl_hist=self.cost_ensemble_ctrl_hist,
                         cost_ensemble_ctrl_sigma_hist=self.cost_ensemble_ctrl_sigma_hist,
                         best_cost_hist=self.best_cost_hist,
                         ctrlVarCoeff_mu_hist=self.ctrlVarCoeff_mu_hist,
                         ctrlVarCoeff_sigma_hist=self.ctrlVarCoeff_sigma_hist,
                         corr_f_hist=self.corr_f_hist,
                         x_bounded_norm_hist=self.x_bounded_norm_hist,
                         eta_sched_hist=self.eta_sched_hist,
                         sigma_ensemble_hist=self.sigma_ensemble_hist,
                         n_iter=self.n_iter,
                         adam_iter=adam_iter,
                         x_bounded_norm_ref=x_bounded_norm_ref,
                         )
                         
                np.savez(self.output_filename + "_GEGD_density_hist",
                         x_hist=self.x_hist,
                         x_latent_hist=self.x_latent_hist,
                         best_x_hist=self.best_x_hist,
                         corr_mu_hist=self.corr_mu_hist,
                         corr_sigma_hist=self.corr_sigma_hist,
                         x_bounded=x_bounded,
                         jac_mean=jac_mean,
                         jac_var=jac_var,
                         )
            
            else:
                np.savez(self.output_filename + "_GEGD_results",
                         N_high_fidelity_hist=self.N_high_fidelity_hist,
                         N_low_fidelity_hist=self.N_low_fidelity_hist,
                         var_reduction_hist=self.var_reduction_hist,
                         N_eff_hist=self.N_eff_hist,
                         cost_ensemble_hist=self.cost_ensemble_hist,
                         cost_ensemble_sigma_hist=self.cost_ensemble_sigma_hist,
                         cost_ensemble_ctrl_hist=self.cost_ensemble_ctrl_hist,
                         cost_ensemble_ctrl_sigma_hist=self.cost_ensemble_ctrl_sigma_hist,
                         best_cost_hist=self.best_cost_hist,
                         ctrlVarCoeff_mu_hist=self.ctrlVarCoeff_mu_hist,
                         ctrlVarCoeff_sigma_hist=self.ctrlVarCoeff_sigma_hist,
                         corr_f_hist=self.corr_f_hist,
                         x_bounded_norm_hist=self.x_bounded_norm_hist,
                         eta_sched_hist=self.eta_sched_hist,
                         n_iter=self.n_iter,
                         adam_iter=adam_iter,
                         x_bounded_norm_ref=x_bounded_norm_ref,
                         )
                         
                np.savez(self.output_filename + "_GEGD_density_hist",
                         x_hist=self.x_hist,
                         sigma_ensemble_hist=self.sigma_ensemble_hist,
                         sigma_fp_hist=self.sigma_fp_hist,
                         x_latent_hist=self.x_latent_hist,
                         best_x_hist=self.best_x_hist,
                         corr_mu_hist=self.corr_mu_hist,
                         corr_sigma_hist=self.corr_sigma_hist,
                         x_bounded=x_bounded,
                         jac_mean=jac_mean,
                         jac_var=jac_var,
                         )

    def sample_cost_fct_landscape(self, output_filename, Nensemble, r_CV, dx1, dx2, Ngrid1interp, Ngrid2interp, Ngrid1, Ngrid2, read_save_data=False, zoom=1):
        if comm.rank == 0:
            print('### Sampling Cost Function Landscape', flush=True)
        
        if read_save_data:
            with np.load(output_filename + "_GEGD_landscape.npz") as data:
                cost_STE = data['cost_STE']
                x_STE = data['x_STE']
                jac_STE = data['jac_STE']
                cost_ensemble = data['cost_ensemble']
                jac_ensemble_mu = data['jac_ensemble_mu']
                jac_NES_mu = data['jac_NES_mu']
                jac_ensemble_sigma = data['jac_ensemble_sigma']
                jac_NES_sigma = data['jac_NES_sigma']
            n1_start = np.min(np.argwhere(cost_ensemble[:,0]==0))
            
        else:
            cost_STE = np.zeros((Ngrid1, Ngrid2))
            x_STE = np.zeros((Ngrid1, Ngrid2, self.Nx, self.Ny))
            jac_STE = np.zeros((Ngrid1, Ngrid2, self.Ndim))
            cost_ensemble = np.zeros((Ngrid1, Ngrid2))
            jac_ensemble_mu = np.zeros((Ngrid1, Ngrid2, self.Ndim))
            jac_ensemble_sigma = np.zeros((Ngrid1, Ngrid2, self.Nsigma))
            jac_NES_mu = np.zeros((Ngrid1, Ngrid2, self.Ndim))
            jac_NES_sigma = np.zeros((Ngrid1, Ngrid2, self.Nsigma))
            n1_start = 0
        
        lb = np.zeros(self.Ndim+self.Nsigma)
        lb[:self.Ndim] = -1
        lb[self.Ndim:] = 0.99*self.sigma_ensemble_max/2
        ub = np.zeros(self.Ndim+self.Nsigma)
        ub[:self.Ndim] = 1
        ub[self.Ndim:] = 1.01*self.sigma_ensemble_max/2
        
        sigma_ensemble = self.sigma_ensemble_max/2
        
        if self.Nsigma == 1:
            sigma_fp = sigma_ensemble.copy()
        else:
            sigma_scaled = 2*(sigma_ensemble - lb[self.Ndim:])/(ub[self.Ndim:] - lb[self.Ndim:]) - 1
            sigma_fp_scaled = dtf.filter_and_project(sigma_scaled, self.symmetry, self.periodic, self.Nx, self.Ny, self.brush_size, self.sigma_filter, self.beta_proj_sigma, padding=self.padding)
            sigma_fp = (sigma_fp_scaled + 1)*(ub[self.Ndim:] - lb[self.Ndim:])/2 + lb[self.Ndim:]
        cov, cov_inv = self.construct_cov_matrix(sigma_fp)

        if comm.rank == 0:
            np.savez(output_filename + "_GEGD_landscape", dx1=dx1, dx2=dx2)

        self.Nensemble = Nensemble
        self.r_CV = r_CV
        for n1 in range(n1_start, Ngrid1):
            if comm.rank == 0:
                print('\tN1: ' + str(n1) + ' | N2: ', end='', flush=True)
            rn = n1/(Ngrid1 - 1)/zoom
                
            for n2 in range(Ngrid2):
                if comm.rank == 0:
                    print(str(n2) + ' ', end='', flush=True)
                
                if n2 == Ngrid2 - 1:
                    cost_STE[n1,n2] = cost_STE[n1,0]
                    x_interp = np.linspace(0, 1, Ngrid1interp)
                    y_interp = np.linspace(0, 1, Ngrid2interp)
                    cost_STE_interp = interpn((np.linspace(0, 1, Ngrid1), np.linspace(0, 1, Ngrid2)),
                                              cost_STE,
                                              np.array(list(product(x_interp, y_interp))),
                                              method='nearest').reshape(Ngrid1interp, Ngrid2interp)
                    x_STE[n1,n2,:,:] = x_STE[n1,0,:,:]
                    jac_STE[n1,n2,:] = jac_STE[n1,0,:]
                    
                    cost_ensemble[n1,n2] = cost_ensemble[n1,0]
                    cost_ensemble_interp = interpn((np.linspace(0, 1, Ngrid1), np.linspace(0, 1, Ngrid2)),
                                                   cost_ensemble,
                                                   np.array(list(product(x_interp, y_interp))),
                                                   method='linear').reshape(Ngrid1interp, Ngrid2interp)
                    jac_ensemble_mu[n1,n2,:] = jac_ensemble_mu[n1,0,:]
                    jac_ensemble_sigma[n1,n2,:] = jac_ensemble_sigma[n1,0,:]
                    jac_NES_mu[n1,n2,:] = jac_NES_mu[n1,0,:]
                    jac_NES_sigma[n1,n2,:] = jac_NES_sigma[n1,0,:]
                
                else:
                    th_n = 2*np.pi/(Ngrid2 - 1)*n2
                    xn = rn*(np.cos(th_n)*dx1 + np.sin(th_n)*dx2)
                
                    dx = np.random.multivariate_normal(np.zeros(self.Ndim), cov, size=self.r_CV*self.Nensemble)
                    cost_all, density_all, grad_STE, grad_ensemble = self.ensemble_jacobian(xn, dx, sigma_fp, True)
                        
                    cost_STE[n1,n2] = cost_all[0]
                    x_interp = np.linspace(0, 1, Ngrid1interp)
                    y_interp = np.linspace(0, 1, Ngrid2interp)
                    cost_STE_interp = interpn((np.linspace(0, 1, Ngrid1), np.linspace(0, 1, Ngrid2)),
                                              cost_STE,
                                              np.array(list(product(x_interp, y_interp))),
                                              method='nearest').reshape(Ngrid1interp, Ngrid2interp)
                    x_STE[n1,n2,:,:] = density_all[1].reshape(self.Nx, self.Ny)
                    jac_STE[n1,n2,:] = grad_STE[1]
                    
                    cost_ensemble[n1,n2] = cost_all[1]
                    cost_ensemble_interp = interpn((np.linspace(0, 1, Ngrid1), np.linspace(0, 1, Ngrid2)),
                                                   cost_ensemble,
                                                   np.array(list(product(x_interp, y_interp))),
                                                   method='linear').reshape(Ngrid1interp, Ngrid2interp)
                    jac_ensemble_mu[n1,n2,:] = grad_ensemble[1][:self.Ndim]
                    jac_ensemble_sigma[n1,n2,:] = grad_ensemble[1][self.Ndim:]
                    jac_NES_mu[n1,n2,:] = cov @ jac_ensemble_mu[n1,n2,:]
                    jac_NES_sigma[n1,n2,:] = sigma_fp**2*jac_ensemble_sigma[n1,n2,:]/2
                    jac_NES_sigma[n1,n2,:] *= 2/(ub[self.Ndim:] - lb[self.Ndim:]) # accounts for covariance rescaling
            
            if comm.rank == 0:
                print('', flush=True)
                
                np.savez(output_filename + "_GEGD_landscape",
                         dx1=dx1,
                         dx2=dx2,
                         cost_STE=cost_STE,
                         cost_STE_interp=cost_STE_interp,
                         x_STE=x_STE,
                         jac_STE=jac_STE,
                         cost_ensemble=cost_ensemble,
                         cost_ensemble_interp=cost_ensemble_interp,
                         jac_ensemble_mu=jac_ensemble_mu,
                         jac_ensemble_sigma=jac_ensemble_sigma,
                         jac_NES_mu=jac_NES_mu,
                         jac_NES_sigma=jac_NES_sigma)